from pygame import mixer

from pynput.keyboard import Listener

from keyboardsounds.profile import Profile
from keyboardsounds.audio_manager import AudioManager

_am: AudioManager = None
_volume = 100
_down = []

def on_press(key):
    global _am
    global _volume
    global _down

    if key in _down:
        return

    sound = _am.get_sound(key, action="press")
    if sound is not None:
        clip = mixer.Sound(sound)
        clip.set_volume(float(_volume) / float(100))
        clip.play()

    # Play the sound
    _down.append(key)

def on_release(key):
    global _down
    global _am

    sound = _am.get_sound(key, action="release")
    if sound is not None:
        clip = mixer.Sound(sound)
        clip.set_volume(float(_volume) / float(100))
        clip.play()

    _down = [k for k in _down if k != key]

def run(volume: int, profile: str):
    global _am
    global _volume

    _am = AudioManager(Profile(profile))
    _volume = volume

    if _am.prime_audio_clips():
        mixer.init()
        with Listener(on_press=on_press, on_release=on_release) as listener:
            listener.join()


import os

from keyboardsounds.path_resolver import PathResolver

VALID_PROFILE_TYPES = ["video-extract", "files"]
SUPPORTED_AUDIO_FORMATS = ["wav", "mp3"]
SUPPORTED_VIDEO_FORMATS = ["mp4"]

def validate_profile(path_resolver: PathResolver, data: dict):
    # Validate Required Generic Keys

    required_keys = ["profile", "sources"]
    for required_key in required_keys:
        if required_key not in data:
            raise ValueError(f"Profile is corrupted. Missing '{required_key}' in profile.yaml.")

    # Validate Profile Metadata
    data = validate_meta_data(path_resolver, data)
    name = data["profile"]["name"]

    # Validate Sources
    validate_sources(path_resolver, name, data)

    # Validate Keys
    validate_keys(path_resolver, name, data)

    return data

def validate_meta_data(path_resolver: PathResolver, data: dict) -> str:
    if "name" not in data["profile"]:
        raise ValueError(f"Profile is corrupted. Missing 'name' in profile.yaml.")
    if type(data["profile"]["name"]) != str:
        raise ValueError(f"Profile is corrupted. 'name' in profile.yaml must be a string.")
    name = data["profile"]["name"]
    if "type" in data["profile"]:
        if data["profile"]["type"] not in VALID_PROFILE_TYPES:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'type' in profile.yaml.")
    else:
        data["profile"]["type"] = "files"
    if "author" in data["profile"]:
        if type(data["profile"]["author"]) != str:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'author' in profile.yaml.")
    if "description" in data["profile"]:
        if type(data["profile"]["description"]) != str:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'description' in profile.yaml.")

    if data["profile"]["type"] == "video-extract":
        if "video" not in data["profile"]:
            raise ValueError(f"Profile '{name}' is corrupted. Missing 'video' in profile.yaml.")
        if type(data["profile"]["video"]) != str:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'video' in profile.yaml.")
        if str(data["profile"]["video"])[-3:] not in SUPPORTED_VIDEO_FORMATS:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'video' in profile.yaml.")
        if not os.path.isfile(path_resolver.get_file_path(data["profile"]["video"])):
            raise ValueError(f"Profile '{name}' is corrupted. Missing video file '{data['profile']['video']}' in profile folder.")
    
    return data

def validate_sources(path_resolver: PathResolver, name: str, data: dict):
    if "sources" not in data:
        raise ValueError(f"Profile '{name}' is corrupted. Missing 'sources' in profile.yaml.")
    if type(data["sources"]) != list:
        raise ValueError(f"Profile '{name}' is corrupted. Invalid 'sources' in profile.yaml.")
    for source in data["sources"]:
       validate_source(path_resolver, name, data, source)

def validate_source(path_resolver: PathResolver, name: str, data: dict, source: any):
    if type(source) != dict:
        raise ValueError(f"Profile '{name}' is corrupted. Invalid 'sources' in profile.yaml.")
    if "id" not in source:
        raise ValueError(f"Profile '{name}' is corrupted. Missing 'id' one or more source in profile.yaml.")

    if data["profile"]["type"] == "files":
        if "source" not in source:
            raise ValueError(f"Profile '{name}' is corrupted. Missing 'source' one or more source in profile.yaml.")
        if type(source["source"]) not in [dict, str]:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'source' one or more source in profile.yaml.")
        if type(source["source"]) == dict:
            if not "press" in source["source"]:
                raise ValueError(f"Profile '{name}' is corrupted. Missing 'press' in source object of one or more sources in profile.yaml.")
            if type(source["source"]["press"]) != str:
                raise ValueError(f"Profile '{name}' is corrupted. Invalid 'press' in source object of one or more sources in profile.yaml.")
            if str(source["source"]["press"])[-3:] not in SUPPORTED_AUDIO_FORMATS:
                raise ValueError(f"Profile '{name}' is corrupted. Invalid 'press' in source object of one or more sources in profile.yaml.")
            if not os.path.isfile(path_resolver.get_file_path(source["source"]["press"])):
                raise ValueError(f"Profile '{name}' is corrupted. Missing audio file '{source['source']['press']}' in profile folder.")
            if "release" in source["source"]:
                if type(source["source"]["release"]) != str:
                    raise ValueError(f"Profile '{name}' is corrupted. Invalid 'release' in source object of one or more sources in profile.yaml.")
                if str(source["source"]["release"])[-3:] not in SUPPORTED_AUDIO_FORMATS:
                    raise ValueError(f"Profile '{name}' is corrupted. Invalid 'release' in source object of one or more sources in profile.yaml.")
                if not os.path.isfile(path_resolver.get_file_path(source["source"]["release"])):
                    raise ValueError(f"Profile '{name}' is corrupted. Missing audio file '{source['source']['release']}' in profile folder.")
        elif type(source["source"]) == str:
            if str(source["source"])[-3:] not in SUPPORTED_AUDIO_FORMATS:
                raise ValueError(f"Profile '{name}' is corrupted. Invalid 'source' in one or more sources in profile.yaml.")
            if not os.path.isfile(path_resolver.get_file_path(source["source"])):
                raise ValueError(f"Profile '{name}' is corrupted. Missing audio file '{source['source']}' in profile folder.")
    elif data["profile"]["type"] == "video-extract":
        if "start" not in source:
            raise ValueError(f"Profile '{name}' is corrupted. Missing 'start' one or more source in profile.yaml.")
        if type(source["start"]) != float:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'start' one or more source in profile.yaml.")
        if "end" not in source:
            raise ValueError(f"Profile '{name}' is corrupted. Missing 'end' one or more source in profile.yaml.")
        if type(source["end"]) != float:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'end' one or more source in profile.yaml.")

def validate_keys(path_resolver: PathResolver, name: str, data: dict):
    if "keys" in data:
        if type(data["keys"]) != dict:
            raise ValueError(f"Profile '{name}' is corrupted. Invalid 'keys' in profile.yaml.")
        if "default" in data["keys"]:
            if type(data["keys"]["default"]) not in [str, list]:
                raise ValueError(f"Profile '{name}' is corrupted. Invalid 'default' in keys in profile.yaml.")
            if type(data["keys"]["default"]) == str:
                validate_source_ref(path_resolver, name, data, data["keys"]["default"])
            elif type(data["keys"]["default"]) == list:
                for source_ref in data["keys"]["default"]:
                    validate_source_ref(path_resolver, name, data, source_ref)
        if "other" in data["keys"]:
            if type(data["keys"]["other"]) != list:
                raise ValueError(f"Profile '{name}' is corrupted. Invalid 'other' in keys in profile.yaml.")
            for key in data["keys"]["other"]:
                validate_key(path_resolver, name, data, key)

def validate_key(path_resolver: PathResolver, name: str, data: dict, key: any):
    if type(key) not in [str, list, dict]:
        raise ValueError(f"Profile '{name}' is corrupted. Invalid 'default' in keys in profile.yaml.")
    if "sound" not in key:
        raise ValueError(f"Profile '{name}' is corrupted. Missing 'sound' in one or more keys in profile.yaml.")
    if type(key["sound"]) not in [list, str]:
        raise ValueError(f"Profile '{name}' is corrupted. Invalid 'sound' in one or more keys in profile.yaml.")
    if type(key["sound"]) == list:
        for source_ref in key["sound"]:
            validate_source_ref(path_resolver, name, data, source_ref)
    else:
        validate_source_ref(path_resolver, name, data, key["sound"])
    pass

def validate_source_ref(path_resolver: PathResolver, name: str, data: dict, source_ref: any):
    if type(source_ref) != str:
        raise ValueError(f"Profile '{name}' is corrupted. Invalid 'sound' in one or more keys in profile.yaml.")
    if source_ref is None or source_ref not in [source["id"] if "id" in source else None for source in data["sources"]]:
        raise ValueError(f"Profile '{name}' is corrupted. Invalid 'sound' in one or more keys in profile.yaml.")


import argparse
import os

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'
import pygame

from importlib.metadata import version

from keyboardsounds.daemon_manager import DaemonManager
from keyboardsounds.profile import Profile
from keyboardsounds.root import ROOT

def main():
    LOCK_FILE = f"{ROOT}/.lock"

    # Work around to get pygame to load mp3 files on windows
    # see https://github.com/pygame/pygame/issues/2647
    if os.name == "nt":
        os.add_dll_directory(pygame.__path__[0])

    dm = DaemonManager(LOCK_FILE)
    if dm.capture_daemon_initialization():
        return

    version_number = version("keyboardsounds")

    usage = (
        f"usage: {os.linesep * 2}"
        f"  manage daemon:{os.linesep * 2}"
        f"    %(prog)s start [-v <volume>] [-p <profile>] [-r]{os.linesep}"
        f"    %(prog)s stop{os.linesep}"
        f"    %(prog)s status{os.linesep * 2}"
        f"  manage profiles:{os.linesep * 2}"
        f"    %(prog)s add-profile -z <zipfile>{os.linesep}"
        f"    %(prog)s remove-profile -n <profile>{os.linesep}"
        f"    %(prog)s list-profiles{os.linesep * 2}"
        f"  other:{os.linesep * 2}"
        f"    %(prog)s version{os.linesep}"
    )

    parser = argparse.ArgumentParser(
        prog=f"keyboardsounds",
        usage=argparse.SUPPRESS,
        description=f"Keyboard Sounds v{version_number}{os.linesep * 2}{usage}{os.linesep}",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument("action", choices=["start", "stop", "status", "add-profile", "remove-profile", "list-profiles"], help="The action to perform")
    
    # Start Action
    parser.add_argument("-v", "--volume", type=int, default=100, metavar="volume", help="volume of the sound effects (0-100), default 100")
    parser.add_argument("-p", "--profile", type=str, default="ios", metavar="profile", help="sound profile to use, default 'ios'")

    # Profiles
    parser.add_argument("-n", "--name", type=str, default=None, metavar="name", help="name of the profile remove")
    parser.add_argument("-z", "--zip", type=str, default=None, metavar="file", help="path to the zip file containing the profile to add")

    parser.add_argument("-V", "--version", action="version", version=version_number)

    args = parser.parse_args()

    if args.action == "start":
        status = dm.status()
        if status == "running":
            print("Re-configuring running instance of Keyboard Sounds daemon...")
        elif status == "stale" or status == "free":
            print("Starting Keyboard Sounds daemon...")
        if not dm.try_start(volume=args.volume, profile=args.profile):
            print("Failed to start.")
            return;
        print(f"Started Keyboard Sounds.")
    elif args.action == "stop":
        stopped = dm.try_stop()
        print("Stopped Keyboard Sounds daemon." if stopped else "Failed to stop Keyboard Sounds daemon. Is it running?")
    elif args.action == "status":
        status = dm.status(full=True)
        print(f"{status}")
    elif args.action == "add-profile":
        if args.zip is None:
            print("Please specify a zip file for the profile to add.")
            return
        Profile.add_profile(args.zip)
        return
    elif args.action == "remove-profile":
        if args.name is None:
            print("Please specify a name for the profile to remove.")
            return
        Profile.remove_profile(args.name)
        return
    elif args.action == "list-profiles":
        profiles = [profile.metadata() for profile in Profile.list()]

        names = [profile["name"] for profile in profiles]
        names.append('Name')
        authors = [profile["author"] for profile in profiles]
        authors.append('Author')
        name_len = len(max(names, key=len))
        auth_len = len(max(authors, key=len))

        print(f"{os.linesep} Available profiles{os.linesep}")
        print(f" {'Name'.ljust(name_len)} | {'Author'.ljust(auth_len)} | Description")
        print(f" {'-' * name_len} | {'-' * auth_len} | -----------")
        for profile in profiles:
            print(f" {profile['name'].ljust(name_len)} | {profile['author'].ljust(auth_len)} | {profile['description']}")
        print(os.linesep)

        return
    else:
        parser.print_usage()

if __name__ == "__main__":
    main()



import os
import yaml
import zipfile
import shutil
import tempfile

from keyboardsounds.root import ROOT
from keyboardsounds.path_resolver import PathResolver
from keyboardsounds.profile_validation import validate_profile


class Profile(PathResolver):
    def __init__(self, name: str):
        super().__init__(os.path.join(ROOT, "profiles", name))
        self.name = name
        self.validate()

    def validate(self):
        # Validate File Structure

        if not os.path.isdir(self.root):
            raise ValueError(f"Profile '{self.name}' does not exist")
        if not os.path.isfile(self.get_file_path("profile.yaml")):
            raise ValueError(f"Profile '{self.name}' is corrupted. Missing profile.yaml.")
    
        with open(self.get_file_path('profile.yaml'), "r") as f:
            data = yaml.safe_load(f)
            self.data = validate_profile(self, data)

    def value(self, key: str):
        value = self.data
        map = key.split(".")
        for key in map:
            if key not in value:
                return None
            value = value[key]
        return value

    def metadata(self):
        name = self.data["profile"]["name"]
        author = self.data["profile"]["author"] if "author" in self.data["profile"] else None
        description = self.data["profile"]["description"] if "description" in self.data["profile"] else None
        return {
            "name": name,
            "author": author,
            "description": description
        }

    @classmethod
    def list(cls):
        names = [f.name for f in os.scandir(os.path.join(ROOT, "profiles")) if f.is_dir()]
        return [Profile(name) for name in names]

    @classmethod
    def remove_profile(cls, name: str):
        output_path = os.path.join(ROOT, "profiles", name)
        if os.path.isdir(output_path):
            shutil.rmtree(output_path)
            print("Profile removed.")
        else:
            print("Profile does not exist.")

    @classmethod
    def add_profile(cls, path: str):
        input_path = os.path.abspath(path)
        tmpdir = tempfile.mkdtemp()

        with zipfile.ZipFile(input_path, "r") as zip_ref:
            zip_ref.extractall(tmpdir)

        profile_data = None
        if os.path.isfile(os.path.join(tmpdir, "profile.yaml")):
            with open(os.path.join(tmpdir, "profile.yaml"), "r") as f:
                profile_data = yaml.safe_load(f)

        if not profile_data or "profile" not in profile_data or "name" not in profile_data["profile"]:
            shutil.rmtree(tmpdir)
            raise ValueError("Profile is corrupted. Missing or invalid profile.yaml.")
        name = profile_data["profile"]["name"]

        print(f"Importing profile from '{input_path}'...")
        output_path = os.path.join(ROOT, "profiles", name)

        shutil.move(tmpdir, output_path)        
        try:
            Profile(name)
            print(f"Successfully imported profile '{name}'.")
        except ValueError as e:
            print(f"Profile '{name}' is invalid. Deleting...")
            shutil.rmtree(output_path)
            print(f"Error: {e}")
            return
        

    
    